import pytest

from sneks.engine.config.instantiation import config


def main(test_path: str | None = None) -> int:
    if test_path is not None:
        config.registrar_prefix = test_path
    return pytest.main(["--pyargs", "sneks.engine.validator"])



import itertools
import random
from collections import Counter
from operator import methodcaller
from typing import FrozenSet, List, Set

from sneks.engine.config.instantiation import config
from sneks.engine.core.cell import Cell
from sneks.engine.engine import cells, registrar
from sneks.engine.engine.mover import Mover, NormalizedScore, Score


class State:
    def __init__(self):
        self.cells: Set[Cell] = {
            Cell(r, c)
            for r, c in itertools.product(
                range(config.game.rows), range(config.game.columns)
            )
        }
        self.active_snakes: List[Mover] = []
        self.ended_snakes: List[Mover] = []
        self.steps: int = 0
        self.occupied: set[Cell] = set()

    def reset(self):
        self.steps = 0
        self.active_snakes = []
        self.ended_snakes = []
        self.occupied = set()
        sneks = registrar.get_submissions()
        sneks.sort(key=lambda s: s.name)
        color_index = 0
        color_index_delta = max(len(config.graphics.colors.snake) // len(sneks), 1)
        for snek in sneks:
            self.active_snakes.append(
                Mover(
                    name=snek.name,
                    head=self.get_random_free_cell(),
                    snek=snek.snek,
                    color=config.graphics.colors.snake[color_index],
                )
            )
            color_index = (color_index + color_index_delta) % len(
                config.graphics.colors.snake
            )
        self.set_board()

    def score_sneks_ended(self) -> None:
        occupations = self.count_occupied_cells()
        for snake in itertools.chain(self.active_snakes, self.ended_snakes):
            for cell in snake.cells:
                if (
                    occupations[cell] > 1 and cell is not snake.get_head()
                ):  # second part should be guaranteed
                    snake.ended += 1

    def count_occupied_cells(self) -> Counter:
        values = itertools.chain(self.active_snakes, self.ended_snakes)
        return Counter(itertools.chain(*(s.cells for s in values)))

    def report(self) -> List[NormalizedScore]:
        self.score_sneks_ended()

        scores = [
            s.get_score()
            for s in itertools.chain(self.active_snakes, self.ended_snakes)
        ]
        min_age = min(s.age for s in scores)
        max_age = max(s.age for s in scores)
        min_ended = min(s.ended for s in scores)
        max_ended = max(s.ended for s in scores)

        min_score = Score(name="min", age=min_age, ended=min_ended)
        max_score = Score(
            name="max",
            age=max(min_age + 1, max_age),
            ended=max(min_ended + 1, max_ended),
        )

        normalized = [
            s.normalize(min_score=min_score, max_score=max_score) for s in scores
        ]
        normalized.sort(key=methodcaller("total"), reverse=True)

        return normalized

    def get_random_free_cell(self):
        options = self.cells.difference(self.get_occupied_cells())
        if options:
            return random.choice(tuple(options))
        else:
            return None

    def get_occupied_cells(self, snakes: List[Mover] | None = None) -> FrozenSet[Cell]:
        if snakes is None:
            values = itertools.chain(self.active_snakes, self.ended_snakes)
            return frozenset().union(*(s.cells for s in values))
        else:
            return frozenset().union(*(s.cells for s in snakes))

    def set_board(self):
        for current_snake in self.active_snakes:
            head = current_snake.get_head()

            # build a grid around the head based on the vision range
            grid = {
                Cell(r, c)
                for r, c in itertools.product(
                    range(
                        head.row - config.game.vision_range,
                        head.row + config.game.vision_range,
                    ),
                    range(
                        head.column - config.game.vision_range,
                        head.column + config.game.vision_range,
                    ),
                )
            }

            # set the snek's occupied to occupied cells within grid
            current_snake.snek.occupied = frozenset(
                cells.get_relative_to(cell, head)
                for cell in grid.intersection(self.occupied)
            )

    def should_continue(self, turn_limit):
        return self.steps < turn_limit and self.active_snakes

    def end_snake(self, snake):
        self.active_snakes.remove(snake)
        self.ended_snakes.append(snake)

    def step(self):
        # add previous head to occupied
        self.occupied |= {s.get_head() for s in self.active_snakes}

        # move the heads
        for snake in self.active_snakes:
            snake.move()

        occupations = Counter(s.get_head() for s in self.active_snakes)

        to_end = []
        # determine ended snakes
        for snake in self.active_snakes:
            if snake.get_head() in self.occupied:
                to_end.append(snake)
            elif occupations[snake.get_head()] > 1:
                to_end.append(snake)

        for snake in to_end:
            self.end_snake(snake)

        for snake in self.active_snakes:
            snake.age += 1

        self.set_board()
        self.steps += 1



from sneks.engine.config.instantiation import config
from sneks.engine.core.cell import Cell
from sneks.engine.core.direction import Direction
from sneks.engine.engine import registrar, runner
from sneks.engine.engine.mover import NormalizedScore
from sneks.engine.interface.snek import Snek


def test_basic_functionality() -> None:
    submissions = registrar.get_submissions()
    assert 1 == len(submissions)
    snek: Snek = submissions[0].snek
    snek.occupied = frozenset((Cell(1, 1),))
    assert snek.get_next_direction() in Direction


def test_extended_functionality() -> None:
    assert config.graphics is not None
    config.graphics.display = False
    config.turn_limit = 100
    scores = runner.main()
    assert scores is not None
    assert len(scores) == 10
    for score in scores:
        assert isinstance(score, NormalizedScore)
        assert score.age == 0
        assert score.ended == 0
        assert score.raw.age >= 0
        assert score.raw.ended == 0


def test_multiple_functionality() -> None:
    assert config.graphics is not None
    config.graphics.display = False
    config.registrar_submission_sneks = 100
    scores = runner.main()
    assert scores is not None
    assert len(scores) == 1000
    for score in scores:
        assert isinstance(score, NormalizedScore)
        assert 0 <= score.age <= 1
        assert 0 <= score.ended <= 1
        assert score.raw.age >= 0
        assert score.raw.ended >= 0



from sneks.engine.gui import graphics, recorder


def main():
    """
    Draws the game board with nothing on it and saves the image

    :return:
    """
    painter = graphics.Painter(recorder.Recorder())
    painter.initialize()
    painter.clear()
    painter.draw_boarders()
    painter.draw()



import hashlib
import os
import struct
import sys

from sneks.engine.core.cell import Cell

try:
    import pygame
    from pygame import Surface
except ModuleNotFoundError:
    pygame = object  # type: ignore
    Surface = object  # type: ignore

from sneks.engine.config.instantiation import config
from sneks.engine.gui.recorder import Recorder

assert config.graphics is not None

ROWS = config.game.rows
COLUMNS = config.game.columns
CELL_SIZE = config.graphics.cell_size
PADDING = config.graphics.padding
COLOR_BORDER = config.graphics.colors.border
COLOR_BACKGROUND = config.graphics.colors.background
COLOR_INVALID = config.graphics.colors.invalid

HEIGHT = (2 + ROWS) * CELL_SIZE + ROWS * PADDING
WIDTH = (2 + COLUMNS) * CELL_SIZE + COLUMNS * PADDING


class Painter:
    screen: Surface | None = None

    def __init__(self, recorder: Recorder | None = None):
        self.recorder = recorder

    def initialize(self):
        if config.graphics.headless:
            os.environ["SDL_VIDEODRIVER"] = "dummy"
        pygame.init()
        self.screen = pygame.display.set_mode((WIDTH, HEIGHT))
        pygame.display.set_caption("Sneks on a Toroidal Plane")

    def draw_snake(self, head: Cell, cells, alive, color: tuple[int, int, int]):
        assert self.screen is not None
        surface = pygame.Surface((CELL_SIZE - PADDING, CELL_SIZE - PADDING))
        fill_horizontal = pygame.Surface((PADDING * 2, CELL_SIZE - PADDING))
        fill_vertical = pygame.Surface((CELL_SIZE - PADDING, PADDING * 2))
        surface.fill(color)
        fill_horizontal.fill(color)
        fill_vertical.fill(color)
        previous = None
        for cell in cells:
            rect = surface.get_rect(
                top=CELL_SIZE + PADDING + cell.row * (CELL_SIZE + PADDING),
                left=CELL_SIZE + PADDING + cell.column * (CELL_SIZE + PADDING),
            )
            self.screen.blit(surface, rect)
            if previous is not None:
                dr, dc = cell.row - previous.row, cell.column - previous.column
                if dr == 0:
                    if dc == -1:
                        rect = fill_horizontal.get_rect(
                            top=CELL_SIZE + cell.row * (CELL_SIZE + PADDING) + PADDING,
                            left=CELL_SIZE
                            + previous.column * (CELL_SIZE + PADDING)
                            - PADDING,
                        )
                    else:
                        rect = fill_horizontal.get_rect(
                            top=CELL_SIZE + cell.row * (CELL_SIZE + PADDING) + PADDING,
                            left=CELL_SIZE
                            + cell.column * (CELL_SIZE + PADDING)
                            - PADDING,
                        )
                    self.screen.blit(fill_horizontal, rect)
                else:
                    if dr == -1:
                        rect = fill_vertical.get_rect(
                            top=CELL_SIZE
                            + previous.row * (CELL_SIZE + PADDING)
                            - PADDING,
                            left=CELL_SIZE
                            + cell.column * (CELL_SIZE + PADDING)
                            + PADDING,
                        )
                    else:
                        rect = fill_vertical.get_rect(
                            top=CELL_SIZE + cell.row * (CELL_SIZE + PADDING) - PADDING,
                            left=CELL_SIZE
                            + cell.column * (CELL_SIZE + PADDING)
                            + PADDING,
                        )
                    self.screen.blit(fill_vertical, rect)
            previous = cell
        if alive:
            surface.fill(
                struct.unpack(
                    "BBB", hashlib.md5(struct.pack("BBB", *color)).digest()[-3:]
                )
            )
            rect = surface.get_rect(
                top=CELL_SIZE + PADDING + head.row * (CELL_SIZE + PADDING),
                left=CELL_SIZE + PADDING + head.column * (CELL_SIZE + PADDING),
            )
            self.screen.blit(surface, rect)

    def draw_ended_head(self, head: Cell):
        assert self.screen is not None
        surface = pygame.Surface((CELL_SIZE - PADDING, CELL_SIZE - PADDING))
        surface.fill(COLOR_INVALID)
        rect = surface.get_rect(
            top=CELL_SIZE + PADDING + head.row * (CELL_SIZE + PADDING),
            left=CELL_SIZE + PADDING + head.column * (CELL_SIZE + PADDING),
        )
        self.screen.blit(surface, rect)

    def clear(self):
        self.screen.fill(COLOR_BACKGROUND)

    def draw_boarders(self):
        top = (0, 0, WIDTH, CELL_SIZE - PADDING)
        bottom = (0, HEIGHT - CELL_SIZE + PADDING, WIDTH, HEIGHT)
        left = (0, CELL_SIZE - PADDING, CELL_SIZE - PADDING, HEIGHT)
        right = (WIDTH - CELL_SIZE + PADDING, 0, WIDTH, HEIGHT)

        for rect in (top, bottom, left, right):
            pygame.draw.rect(self.screen, COLOR_BORDER, rect)

    def draw(self):
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                sys.exit()
        pygame.display.flip()
        if self.recorder:
            self.recorder.record_frame(self.screen)
        self.step_delay()

    def step_delay(self):
        if not config.graphics.headless:
            if config.graphics.step_keypress_wait:
                self.wait_for_keypress()
            pygame.time.delay(config.graphics.step_delay)

    def end_delay(self):
        if not config.graphics.headless:
            if config.graphics.end_keypress_wait:
                self.wait_for_keypress()
            pygame.time.delay(config.graphics.end_delay)

    @staticmethod
    def wait_for_keypress():
        while True:
            # allow the key to be held instead of waiting for each step
            if any(pygame.key.get_pressed()):
                break
            event = pygame.event.wait()
            if event.type == pygame.QUIT:
                pygame.quit()
                sys.exit()
            elif event.type == pygame.KEYDOWN:
                break



import os
import pathlib
import uuid

try:
    import pygame.image
    from pygame import Surface
except ModuleNotFoundError:
    pygame = object()  # type: ignore

from sneks.engine.config.instantiation import config


class Recorder:
    def __init__(self):
        self.identifier = uuid.uuid4()
        self.i = 0
        self.prefix = pathlib.Path(config.graphics.record_prefix)
        self.prefix.mkdir(exist_ok=True)
        (self.prefix / "pics").mkdir(exist_ok=True)
        (self.prefix / "movies").mkdir(exist_ok=True)

    def reset(self):
        self.identifier = uuid.uuid4()
        self.i = 0

    def record_frame(self, screen: Surface):
        pygame.image.save(
            screen,
            str(self.prefix / "pics" / f"pic_{self.identifier}_{self.i:04d}.png"),
        )
        self.i += 1

    def animate_game(self):
        import moviepy.video.io.ImageSequenceClip  # type: ignore

        images = sorted(
            [str(i) for i in self.prefix.glob(f"pics/pic_{self.identifier}_*.png")]
        )

        clip = moviepy.video.io.ImageSequenceClip.ImageSequenceClip(images, fps=24)
        clip.write_videofile(
            str(self.prefix / "movies" / f"game_{self.identifier}.mp4")
        )

        for image in images:
            os.remove(image)



import pathlib

from sneks.engine.config.instantiation import config
from sneks.engine.engine import registrar
from sneks.engine.interface.snek import Snek


def test_import():
    prefix = pathlib.Path(config.registrar_prefix)
    submission_files = registrar.get_submission_files(prefix)
    print(submission_files)
    assert 1 == len(submission_files)
    name = registrar.get_submission_name(submission_files[0])
    assert pathlib.Path(prefix, name, "submission.py").exists()
    registrar.load_module(prefix)

    submissions = registrar.get_submissions()
    assert 1 == len(submissions)
    assert name == submissions[0].name


def test_class_exists():
    _, snek = registrar.get_custom_snek(pathlib.Path(config.registrar_prefix))
    assert issubclass(snek, Snek)



from dataclasses import dataclass
from typing import Tuple

from sneks.engine.core.cell import Cell
from sneks.engine.engine import cells
from sneks.engine.interface.snek import Snek


@dataclass(frozen=True)
class NormalizedScore:
    age: float
    ended: float
    raw: "Score"

    def total(self) -> float:
        return self.age + self.ended

    def __repr__(self):
        return (
            f"age': {self.age:.4f}, ended': {self.ended:.4f}, "
            f"age: {self.raw.age:4d}, ended: {self.raw.ended:2d}, "
            f"name: {self.raw.name}"
        )


@dataclass(frozen=True)
class Score:
    name: str
    age: int
    ended: int

    def normalize(self, min_score: "Score", max_score: "Score") -> NormalizedScore:
        return NormalizedScore(
            age=(self.age - min_score.age) / (max_score.age - min_score.age),
            ended=(self.ended - min_score.ended) / (max_score.ended - min_score.ended),
            raw=self,
        )


class Mover:
    def __init__(self, name: str, head: Cell, snek: Snek, color: Tuple[int, int, int]):
        self.name = name
        self.head = head
        self.cells = {head}
        self.body = [head]
        self.snek = snek
        self.age = 0
        self.color = color
        self.ended = 0

    def get_head(self) -> Cell:
        return self.head

    def move(self):
        next_direction = self.snek.get_next_direction()
        next_head = cells.get_absolute_neighbor(self.get_head(), next_direction)
        self.cells.add(next_head)
        self.body.append(next_head)
        self.head = next_head

    def get_score(self) -> Score:
        return Score(name=self.name, age=self.age, ended=self.ended)



from typing import List, Optional

from sneks.engine.config.instantiation import config
from sneks.engine.engine.mover import NormalizedScore
from sneks.engine.engine.state import State


def demo() -> None:
    config.graphics.display = True
    # config.game.rows = 360
    # config.game.columns = 540
    # config.graphics.cell_size = 2
    config.graphics.step_delay = 0
    config.runs = 100
    config.turn_limit = 10000
    config.registrar_submission_sneks = 100

    main()


def main() -> Optional[List[NormalizedScore]]:
    # import cProfile
    # import pstats
    #
    # pr = cProfile.Profile()
    # pr.enable()

    return main2()

    # pr.disable()
    # stats = pstats.Stats(pr)
    # stats.sort_stats("tottime").print_stats(20)


def main2() -> Optional[List[NormalizedScore]]:
    runs = 0
    state = State()
    state.reset()
    if config.graphics.display:
        from sneks.engine.gui.graphics import Painter
        from sneks.engine.gui.recorder import Recorder

        recorder = None
        if config.graphics.record:
            recorder = Recorder()
        painter = Painter(recorder=recorder)
        painter.initialize()
        while runs < config.runs:
            painter.clear()
            painter.draw_boarders()
            for snake in state.active_snakes:
                painter.draw_snake(snake.head, snake.body, True, snake.color)
            for snake in state.ended_snakes:
                painter.draw_snake(snake.head, snake.body, False, snake.color)
            for snake in state.ended_snakes:
                painter.draw_ended_head(snake.head)
            painter.draw()
            if state.should_continue(config.turn_limit):
                state.step()
            else:
                print(f"Run complete: {runs}")
                if recorder is not None:
                    recorder.animate_game()
                    recorder.reset()
                normalized = state.report()
                for s in normalized:
                    print(f"{s.total():.4f} {s}")
                painter.end_delay()
                runs += 1
                state.reset()
        return None
    else:
        scores = []
        while runs < config.runs:
            if state.should_continue(config.turn_limit):
                state.step()
            else:
                scores += state.report()
                state.reset()
                runs += 1
                if runs % (config.runs / 20) == 0:
                    print("{}% complete".format(100 * runs / config.runs))
        return scores


if __name__ == "__main__":
    main()
